/***********************************************************************
Check whether IEEE 754 signed zero is correctly supported.
[09-May-2001]
***********************************************************************/

#include "config.h"
#include "headers.h"

#define PRINTF (void)printf

typedef union
{
    double f;
    unsigned char i[8];
} D;

#if defined(HAVE_LONG_DOUBLE)
typedef union
{
    long double f;
    unsigned char i[16];
} E;
#endif

typedef union
{
    float f;
    unsigned char i[4];
} S;

static const char* expect_signed(unsigned char e);
static const char* expect_unsigned(unsigned char e);

int
main(int argc, char* argv[])
{
    D d;

#if defined(HAVE_LONG_DOUBLE)
    E e;
#endif

    S s;
    int top;

    /******************************************************************/

    PRINTF("Test of zero handling in single-precision IEEE 754 arithmetic:\n");
    PRINTF("\tsizeof(float) = %u\n", sizeof(float));
    d.f = 1.0;
    top = (d.i[0] == 0) ? 3 : 0;

    s.f = 0.0F;
    PRINTF("\t+zero is %s\n", expect_unsigned(s.i[top]));

    s.f = -0.0F;
    PRINTF("\t-zero is %s\n", expect_signed(s.i[top]));

    s.f = 0.0F - 0.0F;
    PRINTF("\t0 - 0 is %s\n", expect_unsigned(s.i[top]));

    s.f = (-0.0F) - 0.0F;
    PRINTF("\t(-0) - (+0) is %s\n", expect_signed(s.i[top]));

    s.f = 0.0F;
    s.f *= 1.0F;
    PRINTF("\t(+1)*(0) is %s\n", expect_unsigned(s.i[top]));

    s.f = 0.0F;
    s.f *= -1.0F;
    PRINTF("\t(-1)*(0) is %s\n", expect_signed(s.i[top]));

    PRINTF("\n\n");

    /******************************************************************/

    PRINTF("Test of zero handling in double-precision IEEE 754 arithmetic:\n");
    PRINTF("\tsizeof(double) = %u\n", sizeof(double));
    d.f = 1.0;
    top = (d.i[0] == 0) ? 7 : 0;

    d.f = 0.0;
    PRINTF("\t+zero is %s\n", expect_unsigned(d.i[top]));

    d.f = -0.0;
    PRINTF("\t-zero is %s\n", expect_signed(d.i[top]));

    d.f = 0.0 - 0.0;
    PRINTF("\t0 - 0 is %s\n", expect_unsigned(d.i[top]));

    d.f = (-0.0) - 0.0;
    PRINTF("\t(-0) - (+0) is %s\n", expect_signed(d.i[top]));

    d.f = 0.0;
    d.f *= 1.0;
    PRINTF("\t(+1)*(0) is %s\n", expect_unsigned(d.i[top]));

    d.f = 0.0;
    d.f *= -1.0;
    PRINTF("\t(-1)*(0) is %s\n", expect_signed(d.i[top]));

    PRINTF("\n\n");

    /******************************************************************/

#if defined(HAVE_LONG_DOUBLE)
    PRINTF("Test of zero handling in extended-precision IEEE 754 arithmetic:\n");
    PRINTF("\tsizeof(long double) = %u\n", sizeof(long double));
    e.f = 1.0L;
    if (sizeof(long double) == 16u)
	top = (e.i[0] == 0) ? 15 : 0;
    else if (sizeof(long double) == 12u)
	top = (e.i[0] == 0) ? 9 : 0;
    else if (sizeof(long double) == 10u)
	top = (e.i[0] == 0) ? 9 : 0;
    else if (sizeof(long double) == 8u)
	top = (e.i[0] == 0) ? 7 : 0;
    else
    {
	printf("WARNING: unexpected size of extended-precision data type\n");
	top = 0;
    }

    memset(&e,0,sizeof(e));

    e.f = 0.0L;
    PRINTF("\t+zero is %s\n", expect_unsigned(e.i[top]));

    e.f = -0.0L;
    PRINTF("\t-zero is %s\n", expect_signed(e.i[top]));

    e.f = 0.0L - 0.0L;
    PRINTF("\t0 - 0 is %s\n", expect_unsigned(e.i[top]));

    e.f = (-0.0L) - 0.0L;
    PRINTF("\t(-0) - (+0) is %s\n", expect_signed(e.i[top]));

    e.f = 0.0L;
    e.f *= 1.0L;
    PRINTF("\t(+1)*(0) is %s\n", expect_unsigned(e.i[top]));

    e.f = 0.0L;
    e.f *= -1.0L;
    PRINTF("\t(-1)*(0) is %s\n", expect_signed(e.i[top]));

    /******************************************************************/

    PRINTF("\n\n");

#endif

    return (EXIT_SUCCESS);
}

static const char*
expect_signed(unsigned char e)
{
    return ((e == 0x80) ? "  signed\t\tOK" : "unsigned\t\tWRONG");
}

static const char*
expect_unsigned(unsigned char e)
{
    return ((e == 0x00) ? "unsigned\t\tOK" : "  signed\t\tWRONG");
}
